#include <stdarg.h>
#include <libusb-1.0/libusb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <stdint.h>
#include "littlelib.h"

#ifndef ABS
#define ABS(x) ((x) < 0 ? -(x) : (x))
#endif
#ifndef MIN
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#endif

#define INTERFACE               0
#define REFRESH                 200000          /* Poll no more than once every 200ms */

#define ACTION_GETSTATE         1
#define ACTION_SETSTATE         2

#define USB_SHARED_VENDOR       0x16C0
#define USB_SHARED_PRODUCT      0x05DF
#define USB_VENDOR              "captainunlikely.com"
#define USB_PRODUCT             "LittleUSB"

static libusb_context *usbcontext;
static int (*log)(const char *format, ...);

struct little_struct {
    libusb_device_handle *handle;
    struct timeval lastAccess;
    uint16_t usbaddress;
    uint8_t data;
};

//---------------------------------------------------------------------
void little_setlog(int (*printf)(const char *format, ...)) {
    log = printf;
}

static char *usb_error_name(int val) {
#if HAS_LIBUSB_ERROR_NAME == 1
    return libusb_error_name(val);
#else
    static char buf[8];
    snprintf(buf, sizeof(buf), "%d", val);
    return buf;
#endif
}

void little_init() {
    int ret;
    if ((ret=libusb_init(&usbcontext)) < 0) {
        log("culib: init failed: %s\n", usb_error_name(ret));
        return;
    }
    libusb_set_debug(usbcontext, 3);
    little_setlog(&printf);
}

void little_resetall() {
    libusb_device **list;
    libusb_device *usbdevice;
    struct libusb_device_descriptor desc;
    libusb_device_handle *handle;
    int ret;

    ssize_t numdevices = libusb_get_device_list(usbcontext, &list);
    for (ssize_t i=0;i<numdevices;i++) {
        usbdevice = list[i];
        uint16_t address = (libusb_get_bus_number(usbdevice)<<8) | libusb_get_device_address(usbdevice);
        if ((ret=libusb_get_device_descriptor(usbdevice, &desc)) == 0) {
            if (desc.idVendor == USB_SHARED_VENDOR && desc.idProduct == USB_SHARED_PRODUCT) {
                if ((ret=libusb_open(usbdevice, &handle)) == 0) {
                    if ((ret=libusb_reset_device(handle)) == 0) {
                        libusb_close(handle);
                        log("culib: reset 0x%04x\n", address);
                    } else {
                        log("culib: reset failed for 0x%04x: %s \n", address, usb_error_name(ret));
                    }
                } else {
                    log("culib: open failed for 0x%04x: %s \n", address, usb_error_name(ret));
                }
            }
        } else {
            log("culib: can't get descriptor for 0x%04x: %s\n", address, usb_error_name(ret));
        }
    }
    libusb_free_device_list(list, 1);
}

size_t little_list(uint16_t *out, size_t max) {
    size_t count = 0;
    libusb_device **list;
    libusb_device *usbdevice;
    struct libusb_device_descriptor desc;
    libusb_device_handle *handle;
    int ret;
    char buf1[30], buf2[30];
    memset(buf1, 0, sizeof(buf1));
    memset(buf2, 0, sizeof(buf2));

    ssize_t numdevices = libusb_get_device_list(usbcontext, &list);
    for (ssize_t i=0;i<numdevices && count < max;i++) {
        usbdevice = list[i];
        uint16_t address = (libusb_get_bus_number(usbdevice)<<8) | libusb_get_device_address(usbdevice);
        if ((ret=libusb_get_device_descriptor(usbdevice, &desc)) == 0) {
            if (desc.idVendor == USB_SHARED_VENDOR && desc.idProduct == USB_SHARED_PRODUCT) {
                if ((ret=libusb_open(usbdevice, &handle)) == 0) {
#ifdef CLAIMINTERFACE
                    if ((ret=libusb_claim_interface(handle, INTERFACE)) == 0) {
#endif
                        if ((ret=libusb_get_string_descriptor_ascii(handle, desc.iManufacturer, (unsigned char*)buf1, 30)) > 0 && strncmp(buf1, USB_VENDOR, ret) == 0 && (ret=libusb_get_string_descriptor_ascii(handle, desc.iProduct, (unsigned char*)buf2, 30)) > 0 && strncmp(buf2, USB_PRODUCT, ret) == 0) {
                            out[count++] = address;
                        } else if (ret < 0) {
                            log("culib: list: Can't read name for 0x%04x: %s (%s/%s)\n", address, usb_error_name(ret), buf1, buf2);
                            if ((ret=libusb_reset_device(handle)) == 0) {
                                log("culib: reset 0x%04x\n", address);
                            } else {
                                log("culib: reset failed for 0x%04x: %s \n", address, usb_error_name(ret));
                            }
                        }
#ifdef CLAIMINTERFACE
                        if ((ret=libusb_release_interface(handle, INTERFACE)) < 0) {
                            log("culib: list: Can't free interface for 0x%04x: %s\n", address, usb_error_name(ret));
                        }
#endif
                        libusb_close(handle);
#ifdef CLAIMINTERFACE
                    } else {
                        log("culib: list: Can't claim interface for 0x%04x: %s\n", address, usb_error_name(ret));
                        libusb_close(handle);
                    }
#endif
                } else {
                    log("culib: list: Can't open 0x%04x: %s\n", address, usb_error_name(ret));
                }
            }
        } else {
            log("culib: list: Can't get descriptor for 0x%04x: %s\n", address, usb_error_name(ret));
        }
    }
    libusb_free_device_list(list, 1);
    return count;
}

void little_shutdown() {
    libusb_exit(usbcontext);
}

void little_close(little_t s) {
    if (s->handle) {
#ifdef CLAIMINTERFACE
        int ret;
        if ((ret=libusb_release_interface(s->handle, INTERFACE)) != 0) {
            log("culib: Can't release interface for 0x%04x: %s\n", s->usbaddress, usb_error_name(ret));
        }
#endif
        libusb_close(s->handle);
        s->handle = NULL;
    }
}

void little_free(little_t s) {
    if (s->handle) {
        little_close(s);
    }
    free(s);
}

little_t little_open(uint16_t usbaddress) {
    libusb_device **list;
    libusb_device *usbdevice;
    little_t cu = NULL;
    struct libusb_device_descriptor desc;
    libusb_device_handle *handle;
    char buf1[30], buf2[30];
    int ret, found = 0;

    memset(buf1, 0, sizeof(buf1));
    memset(buf2, 0, sizeof(buf2));
    ssize_t numdevices = libusb_get_device_list(usbcontext, &list);
    for (ssize_t i=0;i<numdevices;i++) {
        usbdevice = list[i];
        uint16_t address = (libusb_get_bus_number(usbdevice)<<8) | libusb_get_device_address(usbdevice);
        if (address == usbaddress) {
            if ((ret=libusb_get_device_descriptor(usbdevice, &desc)) == 0) {
                if ((ret=libusb_open(usbdevice, &handle)) == 0) {
#ifdef CLAIMINTERFACE
                    if ((ret=libusb_claim_interface(handle, INTERFACE)) == 0) {
#endif
                        if ((ret=libusb_get_string_descriptor_ascii(handle, desc.iManufacturer, (unsigned char*)buf1, 30)) > 0 && strncmp(buf1, USB_VENDOR, ret) == 0 && (ret=libusb_get_string_descriptor_ascii(handle, desc.iProduct, (unsigned char*)buf2, 30)) > 0 && strncmp(buf2, USB_PRODUCT, ret) == 0) {
                            cu = (little_t )malloc(sizeof(struct little_struct));
                            if (!cu) {
                                perror("malloc failed!");
                                exit(1);
                            }
                            memset(cu, 0, sizeof(struct little_struct));
                            cu->usbaddress = usbaddress;
                            cu->handle = handle;
                        } else if (ret < 0) {
                            log("culib: Can't read strings from USB device 0x%04x: %s (%s/%s)\n", usbaddress, usb_error_name(ret), buf1, buf2);
                            if ((ret=libusb_reset_device(handle)) == 0) {
                                log("culib: reset 0x%04x\n", address);
                            } else {
                                log("culib: reset failed for 0x%04x: %s \n", address, usb_error_name(ret));
                            }
                            libusb_close(handle);
                        } else {
                            libusb_close(handle);
                            log("culib: Incorrect product for USB device 0x%04x: %s/%s\n", usbaddress, buf1, buf2);
                        }
#ifdef CLAIMINTERFACE
                    } else {
                        libusb_close(handle);
                        log("culib: Can't claim USB device 0x%04x: %s\n", usbaddress, usb_error_name(ret));
                    }
#endif
                } else {
                    log("culib: Can't open USB device 0x%04x: %s\n", usbaddress, usb_error_name(ret));
                }
            } else {
                log("culib: Can't open USB device descriptor 0x%04x: %s\n", usbaddress, usb_error_name(ret));
            }
            found = 1;
            break;
        }
    }
    if (!found) {
        log("culib: Can't find LittleUSB device 0x%04x\n", usbaddress);
    }
    libusb_free_device_list(list, 1);
    return cu;
}

static int little_action(little_t cu, int action, int arg) {
    if (!cu->handle) {
        log("culib: USB Device 0x%04x closed\n", cu->usbaddress);
        return -1;
    }
    int ret;
    if ((ret=libusb_control_transfer(cu->handle, LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE | LIBUSB_ENDPOINT_OUT, action, arg, 0, NULL, 0, 500)) < 0) {
        if (ret == LIBUSB_ERROR_NO_DEVICE) {
            little_close(cu);
        } else {
            log("culib: Error running action on 0x%04x: %s", cu->usbaddress, usb_error_name(ret));
        }
    }
    return ret;
}

static int little_update(little_t cu) {
    if (!cu->handle) {
        log("culib: USB Device 0x%04x closed\n", cu->usbaddress);
        return -1;
    }
    struct timeval now;
    int num = 0;

    gettimeofday(&now, NULL);
    if ((now.tv_sec - cu->lastAccess.tv_sec)*1000000 + now.tv_usec - cu->lastAccess.tv_usec > REFRESH) {
        num = libusb_control_transfer(cu->handle, LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE | LIBUSB_ENDPOINT_IN, ACTION_GETSTATE, 0, 0, &cu->data, 1, 500);
        if (num == LIBUSB_ERROR_NO_DEVICE) {
            little_close(cu);
            return num;
        } else if (num < 0) {
            log("culib: Error running update on 0x%04x: %s", cu->usbaddress, usb_error_name(num));
            return num;
        } else if (num == 1) {
            memcpy(&cu->lastAccess, &now, sizeof(now));
        } else {
            log("culib: only %d bytes received from 0x%04x, expecting 1\n", num, cu->usbaddress);
        }
    }
    return 0;
}

int little_get(little_t s) {
    if (little_update(s) == 0) {
        return s->data;
    }
    return -1;
}

int little_set(little_t s, int value) {
    s->data = value;
    return little_action(s, ACTION_SETSTATE, s->data);
}
